# -*- coding: utf-8 -*-
#
# Lithoxyl documentation build configuration file, created by
# sphinx-quickstart on Sun Apr 12 16:44:20 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.

import os
import sys
import sphinx

CUR_PATH = os.path.dirname(os.path.abspath(__file__))
PROJECT_PATH = os.path.abspath(CUR_PATH + '/../')
PACKAGE_PATH = os.path.abspath(CUR_PATH + '/../lithoxyl/')
sys.path.insert(0, PROJECT_PATH)
sys.path.insert(0, PACKAGE_PATH)

# -- General configuration ------------------------------------------------

needs_sphinx = '1.2'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]

# Read the Docs is version 1.2 as of writing
if sphinx.version_info[:2] < (1, 3):
    extensions.append('sphinxcontrib.napoleon')
else:
    extensions.append('sphinx.ext.napoleon')


templates_path = ['_templates']
source_suffix = '.rst'

master_doc = 'index'

# General information about the project.
project = u'Lithoxyl'
copyright = u'2015, Mahmoud Hashemi'
author = u'Mahmoud Hashemi'

# |version| and |release|
# The short X.Y version.
version = '0.4'
# The full version, including alpha/beta/rc tags.
release = '0.4.2'

if os.name != 'nt':
    today_fmt = '%B %e, %Y'

# Added to the end of all rst files before rendering
rst_epilog = ""

exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

intersphinx_mapping = {'python': ('https://docs.python.org/2.7', None)}

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd: # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

#html_theme_options = {}
#html_theme_path = []
#html_title = None
#html_short_title = None
#html_logo = None
#html_favicon = None
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'Lithoxyldoc'

# -- Options for LaTeX output ---------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  (master_doc, 'Lithoxyl.tex', u'Lithoxyl Documentation',
   u'Mahmoud Hashemi', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'lithoxyl', u'Lithoxyl Documentation',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  (master_doc, 'Lithoxyl', u'Lithoxyl Documentation',
   author, 'Lithoxyl', 'One line description of project.',
   'Miscellaneous'),
]




import sys
import json
from xml.etree import ElementTree

import pytest

from lithoxyl import Logger, AggregateSink
from lithoxyl.utils import wrap_all, unwrap_all

IS_PY3 = sys.version_info[0] == 3
IS_PYPY = '__pypy__' in sys.builtin_module_names


def test_wrap_all_json():
    agg_sink = AggregateSink()
    log = Logger('wrapper_log', sinks=[agg_sink])

    wrap_all(log, 'info', json, skip_exc=True)

    json.loads('{}')

    assert agg_sink.begin_events[0].name == 'json.loads'

    unwrap_all(json)

    json.loads('{}')

    if IS_PYPY:
        # different impl
        assert len(agg_sink.begin_events) == 1
    else:
        assert len(agg_sink.begin_events) == 3

    return


@pytest.mark.skipif(IS_PY3, reason='no old-style classes and elementree is a builtin type/cannot setattr on it')
def test_wrap_all_element_tree():
    log = Logger('test', sinks=[])

    # test old-style class wrapping / unwrapping
    wrap_all(log, target=ElementTree)
    unwrap_all(ElementTree)



# -*- coding: utf-8 -*-

import errno

from lithoxyl import (SensibleSink,
                      SensibleFilter,
                      SensibleFormatter as SF,
                      SensibleMessageFormatter as SMF)
from lithoxyl.emitters import StreamEmitter, AggregateEmitter
from lithoxyl.logger import Logger


fmtr = SF('{status_char}{begin_timestamp} - {event_message}')
strm_emtr = StreamEmitter('stderr')
fltr = SensibleFilter('debug')
aggr_emtr = AggregateEmitter()
strm_sink = SensibleSink(formatter=fmtr, emitter=strm_emtr)
fake_sink = SensibleSink(filters=[fltr], formatter=fmtr, emitter=aggr_emtr)


def test_sensible_basic():
    log = Logger('test_ss', [strm_sink, fake_sink])

    log.debug('greet').success('hey')
    assert aggr_emtr.get_entry(-1).startswith('s')

    with log.debug('greet') as t:
        log.comment('a_{}_quick', 'comment')
        assert aggr_emtr.get_entry(-1).startswith('#')
        assert 'a_comment_quick' in aggr_emtr.get_entry(-1)
        t.success('hello')
        t.warn("everything ok?")

    assert aggr_emtr.get_entry(-1).startswith('S')

    with log.debug('greet') as t:
        t.failure('bye')
    assert aggr_emtr.get_entry(-1).startswith('F')

    try:
        with log.debug('greet') as t:
            raise ZeroDivisionError('narwhalbaconderp')
    except Exception:
        pass

    assert aggr_emtr.get_entry(-1).startswith('E')
    assert 'limit=' in repr(aggr_emtr)
    assert aggr_emtr.get_entries()
    aggr_emtr.clear()
    assert not aggr_emtr.get_entries()


def test_bad_encoding():
    try:
        StreamEmitter('stderr', encoding='nope')
    except LookupError:
        assert True
    else:
        assert False


def test_bad_encoding_error_fallback():
    try:
        StreamEmitter('stderr', errors='badvalue')
    except LookupError:
        assert True
    else:
        assert False


def _test_exception():
    _tmpl = ('{iso_end} - {exc_type}: {exc_message}'
             ' - {func_name}:{line_number} - {exc_tb_list}')
    sink = SensibleSink(SF(_tmpl),
                        StreamEmitter('stderr'),
                        filters=[SensibleFilter(exception=False)])
    logger = Logger('excelsilog', [sink])
    with logger.info('A for Effort', reraise=False) as tr:
        print(tr)
        raise ValueError('E for Exception')
    return


def test_stale_stream(tmpdir):
    # make mock filestream with write/flush that goes stale after 100 writes
    # create logger with stream emitter to mocked file stream

    class StalewardFile(object):
        def __init__(self, wrapped):
            self._write_count = 0
            self.wrapped = wrapped

        def write(self, *a, **kw):
            self._write_count += 1
            if self._write_count > 100:
                exc = IOError('stale file handle')
                exc.errno = errno.ESTALE
                self.close()
                raise exc
            return self.wrapped.write(*a, **kw)

        def __getattr__(self, name):
            return getattr(self.wrapped, name)

    file_path = '%s/not_always_fresh.log' % (tmpdir,)
    stale_file_obj = StalewardFile(open(file_path, 'wb'))
    emitter = StreamEmitter(stale_file_obj)

    sink = SensibleSink(SF('{status_char} - {iso_end}'), emitter,
                        filters=[SensibleFilter(success=True)])
    logger = Logger('excelsilog', [sink])

    assert emitter.stream is stale_file_obj

    for i in range(200):
        logger.info('yay').success()

    lines = open(file_path).read().splitlines()
    assert len(lines) == 200
    assert len(lines[0]) == len(lines[-1])
    assert stale_file_obj.closed
    assert emitter.stream is not stale_file_obj



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import sys
import time
import types
import socket
import hashlib
import binascii
from os import getpid

PY3 = sys.version_info[0] == 3

try:
    basestring
    type_types = (type, types.ClassType)
except NameError:
    basestring = str
    type_types = (type,)

class EncodingLookupError(LookupError):
    pass


class ErrorBehaviorLookupError(LookupError):
    pass


def check_encoding_settings(encoding, errors, reraise=True):
    try:
        # then test error-handler
        u''.encode(encoding)
    except LookupError as le:
        if reraise:
            raise EncodingLookupError(le.args[0])
        return False
    try:
        # then test error-handler
        # python2.6 does not support kwargs for encode
        u'\xdd'.encode('ascii', errors)
    except LookupError as le:
        if reraise:
            raise ErrorBehaviorLookupError(le.args[0])
        return False
    except Exception:
        # that ascii encode should never work
        return True


def unwrap(target, attr_name):
    wrapped_func = getattr(target, attr_name)
    try:
        unwrapped_func = wrapped_func.__lithoxyl_wrapped__[3]
    except Exception:
        raise ValueError('%r does not appear to be a wrapped function'
                         % wrapped_func)
    setattr(target, attr_name, unwrapped_func)
    return


def unwrap_all(target=None):
    if target is None or isinstance(target, int):
        target = target or 1
        try:
            target_module_name = sys._getframe(target).f_globals.__name__
            calling_module = sys.modules[target_module_name]
        except Exception:
            raise ValueError('unable to wrap all with target: %r' % target)
        target = calling_module

    unwrapped = set()

    def unwrap_sub_target(sub_target):
        for attr_name in dir(sub_target):
            try:
                val = getattr(sub_target, attr_name)
            except AttributeError:
                continue
            if not callable(val):
                continue
            elif isinstance(val, type_types):
                if id(val) in unwrapped:
                    continue
                unwrapped.add(id(val))
                unwrap_sub_target(val)
            else:
                try:
                    unwrap(sub_target, attr_name)
                except ValueError:
                    continue
        return

    unwrap_sub_target(target)
    return


def wrap_all(logger, level='info', target=None, skip=None,
             label=None, level_map=None, extras=None, depth=1,
             skip_exc=False):
    """
    """
    ret = []
    extras = extras or {}
    level_map = level_map or {}

    if target is None or isinstance(target, int):
        target = target or 1
        try:
            target_module_name = sys._getframe(target).f_globals.__name__
            calling_module = sys.modules[target_module_name]
        except Exception:
            raise ValueError('unable to wrap all with target: %r' % target)
        target = calling_module

    if skip is None:
        skip = '_'
    if isinstance(skip, basestring):
        skip_func = lambda attr_name: skip and attr_name.startswith(skip)
    elif callable(skip):
        skip_func = skip
    else:
        raise ValueError('skip expected string prefix or callable, not %r' %
                         (skip,))
    if skip_exc is True:
        skip_exc = Exception
    elif skip_exc is False:
        skip_exc = ()

    if label is None:
        try:
            label = target.__name__
        except AttributeError:
            label = '(%s@%s)' % (target.__class__.__name__, hex(id(target)))
    label = str(label)

    wrapped = set()

    def wrap_sub_target(sub_target, depth=depth, label=label):
        for attr_name in dir(sub_target):
            if skip_func(attr_name):
                continue
            val = getattr(sub_target, attr_name)
            if not callable(val):
                continue
            elif isinstance(val, type_types):
                # NOTE: it may be possible to get to the same function / method
                # via different attribute paths; we will pick one of them randomly
                # in that case for implementation simplicity
                if depth <= 0 or id(val) in wrapped:
                    continue
                wrapped.add(id(val))
                wrap_sub_target(val, depth - 1, label + '.' + attr_name)
            elif not hasattr(val, '__name__'):
                continue
            else:
                kwargs = dict(extras)

                kwargs['level'] = level_map.get(attr_name, level)
                kwargs['action_name'] = label + '.' + attr_name

                log_wrapper = logger.wrap(**kwargs)
                try:
                    wrapped_func = log_wrapper(val)
                    setattr(sub_target, attr_name, wrapped_func)
                except skip_exc:
                    continue

                ret.append(label + '.' + attr_name)
        return

    wrap_sub_target(target)

    return ret


# I'd love to use UUID.uuid4, but this is 10-20x faster
# 12 bytes (96 bits) means that there's 1 in 2^32 chance of a collision
# after 2^64 messages.
# TODO: fallback to UUID if hashlib isn't available?

def reseed_guid():
    """This is called automatically on fork by the functions below. You
    probably don't need to call this.
    """
    global _PID
    global _GUID_SALT
    global _GUID_START

    try:
        random_hex = os.urandom(4).hex()
    except AttributeError:
        # py2
        random_hex = binascii.hexlify(os.urandom(4))

    _PID = getpid()
    _GUID_SALT = '-'.join([str(getpid()),
                           socket.gethostname() or '<nohostname>',
                           str(time.time()),
                           random_hex])
    _GUID_START = int(hashlib.sha1(_GUID_SALT.encode('utf8')).hexdigest()[:24], 16)

    return


# set in reseed
_PID = None
_GUID_SALT = None
_GUID_START = None


reseed_guid()


def int2hexguid(id_int):
    """
    I'd love to use UUID.uuid4, but this is 20x faster

    sha1 is 20 bytes. 12 bytes (96 bits) means that there's 1 in 2^32
    chance of a collision after 2^64 messages.
    """
    return hashlib.sha1((_GUID_SALT + str(id_int)).encode('utf8')).hexdigest()[:24]


def int2hexguid_seq(id_int):
    """Much like int2hexguid, this function returns a fork-safe,
    globally-unique, 12-character hexadecimal string, based on an
    integer input. Intended for use with the always-incrementing
    action_ids and event_ids.

    This variant is specialized for use in autoincrementing use cases
    like action_id and event_id, as the guids it produces will
    maintain the same sortability characteristics as the original
    ID. (as the input integer increments, so will the output GUID)
    """
    if getpid() != _PID:
        reseed_guid()
    return '%x' % (_GUID_START + id_int)


"""decorator.py is bad because it excessively changes your decorator
API to be reliant on decorator.py's strange aesthetic. A pre-existing
decorator can't easily be migrated, and a decorator.py decorator is
not compatible with functools.wraps.

Function signature propagation is orthogonal to closure usage. The
author of decorator.py seems to find a problem with having a function
inside of a function and/or relying on closures and/or functools.wraps
interface.
"""

# from six, handling the change in py3's raise keyword
if PY3:
    try:
        exec_ = getattr(__builtins__, 'exec')
    except AttributeError:
        exec_ = __builtins__.get('exec')

    def reraise(tp, value, tb=None):
        try:
            if value is None:
                value = tp()
            if value.__traceback__ is not tb:
                raise value.with_traceback(tb)
            raise value
        finally:
            value = None
            tb = None

else:
    def exec_(_code_, _globs_=None, _locs_=None):
        """Execute code in a namespace."""
        if _globs_ is None:
            frame = sys._getframe(1)
            _globs_ = frame.f_globals
            if _locs_ is None:
                _locs_ = frame.f_locals
            del frame
        elif _locs_ is None:
            _locs_ = _globs_
        exec("""exec _code_ in _globs_, _locs_""")

    exec_("""def reraise(tp, value, tb=None):
    try:
        raise tp, value, tb
    finally:
        tb = None
""")




import pytest

from lithoxyl import Logger


def test_wrap():
    logger = Logger('t')

    @logger.wrap('critical')
    def t_func():
        return True

    assert t_func() is True

    @logger.wrap('critical', inject_as='yep')
    def y_func(yep):
        return bool(yep)

    assert y_func() is True

    # try inject_as with an argument that isn't there
    with pytest.raises(ValueError):
        @logger.wrap('critical', inject_as='nope')
        def t_func():
            return False

    return



"""A systematic approach to application instrumentation, including
logging, semantic profiling, and statistics collection. Very
lightweight, very Pythonic.

(c) 2020 by Mahmoud Hashemi.
BSD-licensed, see LICENSE for more details.
"""

import sys
from setuptools import setup, find_packages


__author__ = 'Mahmoud Hashemi'
__version__ = '20.0.0'
__contact__ = 'mahmoud@hatnote.com'
__url__ = 'https://github.com/mahmoud/lithoxyl'
__license__ = 'BSD'

desc = ('A systematic approach to logging, profiling, and statistics'
        'collection. Very lightweight, very Pythonic.')


if sys.version_info < (2,6):
    raise NotImplementedError("Sorry, lithoxyl only supports Python >=2.6")


setup(name='lithoxyl',
      version=__version__,
      description=desc,
      long_description=__doc__,
      author=__author__,
      author_email=__contact__,
      url=__url__,
      packages=find_packages(),
      install_requires=['boltons>=20.0.0'],
      include_package_data=True,
      zip_safe=False,
      license=__license__,
      platforms='any',
      classifiers=[
          'Intended Audience :: Developers',
          'Topic :: System :: Logging',
          'Topic :: Utilities',
          'Topic :: Software Development :: Libraries :: Application Frameworks',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
      ]
)


"""
A brief checklist for release:

* tox
* git commit (if applicable)
* Bump setup.py version off of -dev
* git commit -a -m "bump version for x.y.z release"
* python setup.py sdist bdist_wheel upload
* bump docs/conf.py version
* git commit
* git tag -a x.y.z -m "brief summary"
* write CHANGELOG
* git commit
* bump setup.py version onto n+1 dev
* git commit
* git push

"""



# -*- coding: utf-8 -*-

from lithoxyl import Logger
from lithoxyl.sinks import CounterSink


def test_basic_counter():
    csink = CounterSink()
    log = Logger('ctr_log', [csink])
    log.debug('t').success('est')

    assert csink.counter_map[log]['t'] == 1

    for i in range(1000):
        log.debug('e').success('st')

    assert csink.counter_map[log]['e'] == 1000
    assert csink.counter_map[log].get('t') == 0

    cdict = csink.to_dict()
    assert cdict == {'ctr_log': {'e': 1000,
                                 '__all__': 1001,
                                 '__missing__': 1}}
    return
